package uk.gov.dwp.maze;


import org.junit.Test;
import static org.junit.Assert.*;

public class MazeTest {

    public static String[] MAZE_FIXTURE = new String[]{
            "XXXXXXXXXXXXXXX",
            "X             X",
            "X XXXXXXXXXXX X",
            "X XS        X X",
            "X XXXXXXXXX X X",
            "X XXXXXXXXX X X",
            "X XXXX      X X",
            "X XXXX XXXX X X",
            "X XXXX XXXX X X",
            "X X    XXXXXX X",
            "X X XXXXXXXXX X",
            "X X XXXXXXXXX X",
            "X X         X X",
            "X XXXXXXXXX   X",
            "XFXXXXXXXXXXXXX"};


    @Test
    public void testWhenConstructingAMazeWithIllegalCharacters(){
        try {
            new Maze(new String[]{"XXX$X","X X X","XXXXS"});
            fail("Should have thrown Exception");
        } catch (IllegalArgumentException e) {
            assertEquals("Maze is not properly defined",e.getMessage());
        }
    }

            @Test
    public void testWhenConstructingAMazeFromAStringArrayItMustHaveAnExit() {
        try {
            new Maze(new String[]{"XXXXX","X X X","XXXXS"});
            fail("Should have thrown Exception");
        } catch (IllegalArgumentException e) {
            assertEquals("Maze doesn't have a finish",e.getMessage());
        }
    }

    @Test
    public void testWhenConstructingAMazeFromAStringArrayItMustHaveAStart() {
        try {
            new Maze(new String[]{"XXXXX","X X X","XXXXF"});
            fail("Should have thrown Exception");
        } catch (IllegalArgumentException e) {
            assertEquals("Maze doesn't have a start",e.getMessage());
        }
    }

    @Test
    public void testWhenConstructingAMazeFromAStringArrayCannotHaveMultipleStarts() {
        try {
            new Maze(new String[]{"SXXXF","X X X","XXXXS"});
            fail("Should have thrown Exception");
        } catch (IllegalArgumentException e) {
            assertEquals("Maze with multiple start points not allowed",e.getMessage());
        }
    }

    @Test
    public void testWhenConstructingAMazeFromAStringArrayCannotHaveMultipleExits() {
        try {
            new Maze(new String[]{"FXXXS","X X X","XXXXF"});
            fail("Should have thrown Exception");
        } catch (IllegalArgumentException e) {
            assertEquals("Maze with multiple finish points not allowed",e.getMessage());
        }
    }

    @Test
    public void testCorrectNumberOfSpacesAndWallsReturnedForValidMazes() {
        Maze m1 = new Maze(new String[]{"XXXXS", "X X X", "X X X", "XXXXF"});
        assertEquals(4, m1.getNumberOfSpaces());
        assertEquals(14, m1.getNumberOfWalls());
    }

    @Test
    public void testCorrectDescriptionReturnedForPositionsInSimpleMaze(){
        Maze m1 = new Maze(new String[]{
                "XXXXS",
                "X X X",
                "X X X",
                "XXXXF"});
        assertTrue( m1.isWallAt(0, 0));
        assertTrue( m1.isWallAt(1, 0));
        assertTrue( m1.isWallAt(2, 0));
        assertTrue( m1.isWallAt(3, 0));
        assertFalse(m1.isWallAt(4, 0));

        assertTrue( m1.isWallAt(0, 1));
        assertFalse( m1.isWallAt(1, 1));
        assertTrue( m1.isWallAt(2, 1));
        assertFalse(m1.isWallAt(3, 1));
        assertTrue( m1.isWallAt(4, 1));

        assertTrue( m1.isWallAt(0, 2));
        assertFalse(m1.isWallAt(1, 2));
        assertTrue( m1.isWallAt(2, 2));
        assertFalse(m1.isWallAt(3, 2));
        assertTrue( m1.isWallAt(4, 2));

        assertTrue(m1.isWallAt(0, 3));
        assertTrue(m1.isWallAt(1, 3));
        assertTrue(m1.isWallAt(2, 3));
        assertTrue(m1.isWallAt(3, 3));
        assertTrue(m1.isFinishAt(4,3));
        assertEquals(new Position(4,0),m1.getStart());
        assertEquals(new Position(4,3),m1.getFinish());
    }


    @Test
    public void testCorrectDescriptionReturnedForPositionsInComplexMaze(){
        Maze m = new Maze(MAZE_FIXTURE);
        assertWallOnHorizontalRange(m, 0, 0, 14);
        assertTrue(m.isWallAt(0,1));
        assertTrue(m.isWallAt(14,1));
        assertSpaceOnHorizontalRange(m, 1, 1, 13);
        assertFalse(m.isWallAt(1,2));
        assertFalse(m.isWallAt(13,2));
        assertTrue(m.isWallAt(0, 3));
        assertTrue(m.isWallAt(14, 3));
        assertTrue(m.isWallAt(2, 3));
        assertSpaceOnHorizontalRange(m, 3, 3, 11);
        assertTrue(m.isWallAt(0,4));
        assertFalse(m.isWallAt(1, 4));
        assertWallOnHorizontalRange(m, 4, 2, 10);
        assertFalse(m.isWallAt(11,4));
        assertFalse(m.isWallAt(13,4));
        assertTrue(m.isWallAt(14,4));
        assertTrue(m.isWallAt(0,5));
        assertFalse(m.isWallAt(1, 5));
        assertWallOnHorizontalRange(m, 5, 2, 10);
        assertFalse(m.isWallAt(11,5));
        assertFalse(m.isWallAt(13,5));
        assertTrue(m.isWallAt(14,5));   //ok, that's probably enough asserting on the fixture

    }

    private void assertWallOnHorizontalRange(Maze m,int y, int xStart, int xEnd) {
        for(int x = xStart; x <= xEnd; x++){
            assertEquals("Did not find wall at "+new Position(x,y).toString(),true,m.isWallAt(x,y));
        }
    }

    private void assertSpaceOnHorizontalRange(Maze m,int y, int xStart, int xEnd) {
        for(int x = xStart; x <= xEnd; x++){
            assertEquals("Did not find space at "+new Position(x,y).toString(),false,m.isWallAt(x,y));
        }
    }


}
